//===-- ConnectionFileDescriptor.h ------------------------------*- C++ -*-===//
//
// Part of the LLVM Project, under the Apache License v2.0 with LLVM Exceptions.
// See https://llvm.org/LICENSE.txt for license information.
// SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
//
//===----------------------------------------------------------------------===//

#ifndef LLDB_HOST_CONNECTIONFILEDESCRIPTOR_H
#define LLDB_HOST_CONNECTIONFILEDESCRIPTOR_H

#include "lldb/Host/posix/ConnectionFileDescriptorPosix.h"

#endif
